"""
:babelast:`` AST structures
"""

from jast.util import AST, DecorateMixin, \
    js_var, list_attr, ConvList, \
    AliasFactory, cls, attr, json, convert_as, \
    Factory, partial, is_list

def JS(val): #pylint: disable=invalid-name,too-many-return-statements
    """
    Automatically convert python object to Boolean, Numeric, String or Object representation.
    """
    if val is None:
        return None
    if isinstance(val, AST):
        return val
    if isinstance(val, bool):
        return Boolean(val)
    if isinstance(val, int):
        return Numeric(val)
    if isinstance(val, basestring):
        return String(val)
    if isinstance(val, dict):
        return Object(val.items())
    if is_list(val):
        return Array(val)
    raise ValueError("Cannot resolve %r to AST value" % val)


@cls
class Program(AST):
    """
    :babelast:`programs`
    """

    type = "Program"
    body = list_attr()
    sourceType = attr(default='module')
    interpreter = attr(default=None)
    directives = list_attr()

    ast_keys_order = ['sourceType', 'interpreter', 'body', 'directives']


@cls
class Identifier(AST):
    """
    :babelast:`identifier`
    """
    type = "Identifier"
    name = attr(converter=js_var, default=None)
    definition = attr(default=None)

    @staticmethod
    def auto(val, imp=None):
        """
        Automatically convert value as an AST :class:`.Identifier`
        """
        if hasattr(val, 'identifier'):
            return val.identifier()

        if isinstance(val, AST):
            return val

        if isinstance(val, basestring):
            if "." in val:
                prop = Identifier(val.split(".")[-1])
                obj = Identifier.auto(".".join(val.split(".")[:-1]))
                return Member(object=obj, prop=prop, definition=imp)
            return Identifier(val, definition=imp)

        return val

    def call(self, *args):
        """
        Convert to :class:`.Call` instance.
        """
        return Call(self, args)

    def member(self, arg):
        """
        Convert to :class:`.Member` instance.
        """
        return Member(object=self, prop=Id.auto(arg))


Id = Identifier


@cls
class Extra(AST):
    """
    :babelast:`extra`
    """
    rawValue = attr()
    raw = attr(converter=json.dumps, default=AliasFactory('rawValue'))


@cls
class String(AST):
    """
    :babelast:`string`
    """
    type = "StringLiteral"
    value = attr()
    extra = attr(default=AliasFactory('value'), converter=convert_as(Extra))


@cls
class Numeric(AST):
    """
    :babelast:`numeric`
    """
    type = "NumericLiteral"
    value = attr(converter=int)
    extra = attr(default=AliasFactory('value'), converter=convert_as(Extra))


@cls
class Boolean(AST):
    """
    :babelast:`boolean`
    """
    type = "BooleanLiteral"
    value = attr(converter=bool)


def _import_factory(imp):
    """
    :class:`.Import` specificers converter.
    """
    if is_list(imp):
        return map(ImportSpecifier, imp)
    return [ImportDefaultSpecifier(imp)]

@cls
class Import(AST):
    """
    :babelast:`importdeclaration`
    """
    type = "ImportDeclaration"

    fct = lambda x: x.source.value if hasattr(x, 'source') else None
    specifiers = attr(default=Factory(fct, takes_self=True),
                      converter=_import_factory)
    source = attr(default=None, converter=convert_as(String))

    def id(self, key=None): #pylint: disable=invalid-name
        """
        Shortcut to :py:meth:`.Import.identifier`
        """
        return self.identifier(key)

    def identifier(self, key=None):
        """
        Get import statement specifier as :class:`.Identifier`.
        """
        #pylint: disable=unsubscriptable-object,not-an-iterable
        _id = None
        if key is None:
            assert len(self.specifiers) == 1
            _id = self.specifiers[0].local
        for spec in self.specifiers:
            if spec.local.name == key:
                _id = spec.local
                break
        _id = Identifier.auto(_id.copy())
        _id.definition = self
        return _id


@cls
class ImportSpecifier(AST):
    """
    :babelast:`importspecifier`
    """
    type = "ImportSpecifier"

    imported = attr(converter=Identifier.auto)
    local = attr(default=Factory(lambda x: x.imported, takes_self=True))


def _import_fix(imp):
    auto = Identifier.auto(imp)
    if isinstance(auto, Member):
        return auto.object
    return auto

@cls
class ImportDefaultSpecifier(AST):
    """
    :babelast:`importdefaultspecifier`
    """
    type = "ImportDefaultSpecifier"

    local = attr(converter=_import_fix)


@cls
class ObjectProperty(AST):
    """
    :babelast:`objectproperty`
    """
    type = "ObjectProperty"

    key = attr(default=None, converter=Identifier.auto)
    value = attr(default=None, converter=JS)
    computed = attr(default=False)
    shorthand = attr(default=False)
    method = attr(default=False)

    ast_keys_order = ['type', 'method', 'key', 'computed', 'shorthand', 'value']


def _normalize_dict_to_list(value):
    val = value
    if isinstance(val, dict):
        val = [ObjectProperty(key, JS(v)) for key, v in val.items()]
    return ConvList(val, convert=convert_as(ObjectProperty,
                                            expand_args=True))

@cls
class Object(AST, dict):
    """
    :babelast:`object`
    """
    type = "ObjectExpression"

    properties = attr(converter=_normalize_dict_to_list, default=Factory(
        partial(ConvList,
                convert=convert_as(ObjectProperty, expand_args=True))))


    @property
    def properties_dict(self):
        """
        Convert a js Object to python dict
        """
        #pylint: disable=not-an-iterable
        return dict([(x.key.name, x) for x in self.properties])

    def _property_index(self, key):
        for index, prop in enumerate(self.properties):
            if prop.key == key:
                return index
        return None

    def __getitem__(self, key):
        val = self.properties_dict[key]
        while hasattr(val, 'value'):
            val = getattr(val, 'value')
        return val

    def __setitem__(self, key, value):
        #pylint: disable=unsupported-assignment-operation,no-member
        index = self._property_index(key)
        val = JS(value)
        if not isinstance(val, ObjectProperty):
            val = ObjectProperty(key, val)

        if index:
            self.properties[index] = val
        else:
            self.properties.append(val)

    def __delitem__(self, key):
        #pylint: disable=unsupported-delete-operation
        index = self._property_index(key)
        del self.properties[index]

    def __contains__(self, key):
        return bool(self._property_index(key))



@cls
class NullLiteral(AST):
    """
    :babelast:`nullliteral`
    """
    type = "NullLiteral"

Null = NullLiteral() #pylint: disable=invalid-name


def _normalize_array(val):
    if not is_list(val):
        return JS(val)
    return val

@cls
class Array(AST):
    """
    :babelast:`arrayexpression`
    """
    type = "ArrayExpression"

    elements = list_attr(converter=JS, normalize=_normalize_array)


@cls
class Member(AST):
    """
    :babelast:`memberexpression`
    """
    type = "MemberExpression"
    object = attr(converter=Identifier.auto)
    prop = attr(converter=Identifier.auto)
    computed = attr(default=False)
    definition = attr(default=None)

    ast_map = {
        'prop': 'property'
    }

    def call(self, *args):
        """
        Convert to member Call statement
        """
        return Call(self, args)

@cls
class Var(AST):
    """
    :babelast:`variabledeclaration`
    """
    type = "VariableDeclaration"
    declarations = list_attr()
    kind = attr(default="let")

    @staticmethod
    def simple(identifier, init):
        """
        Common VariableDeclarator AST factory
        """
        return Var(VariableDeclarator(identifier, JS(init)))


@cls
class VariableDeclarator(AST):
    """
    :babelast:`variabledeclarator`
    """
    type = "VariableDeclarator"
    id = attr(converter=Identifier.auto)
    init = attr()


@cls
class Expression(AST):
    """
    :babelast:`expressionstatement`
    """
    type = "ExpressionStatement"
    expression = attr()


@cls
class Block(AST):
    """
    :babelast:`blockstatement`
    """
    type = "BlockStatement"
    body = list_attr(converter=Expression.converter)
    directives = list_attr()


@cls
class Function(AST):
    """
    :babelast:`functionexpression`
    """
    type = "FunctionExpression"

    params = list_attr(converter=Identifier.auto)
    body = attr(default=Factory(Block), converter=Block.converter)
    id = attr(default=None)
    isAsync = attr(default=False)
    generator = attr(default=False)

    ast_map = {
        'isAsync': 'async'
    }


@cls
class NamedFunction(Function):
    """
    :babelast:`functiondeclaration`
    """
    type = "FunctionDeclaration"


@cls
class Call(AST):
    """
    :babelast:`callexpression`
    """
    type = "CallExpression"
    callee = attr(converter=Identifier.auto, default=None)
    arguments = list_attr(converter=JS, default=None)


@cls
class Decorator(AST):
    """
    :babelast:`decorator`
    """
    type = "Decorator"
    expression = attr(converter=JS)


@cls
class CallDecorator(AST):
    """
    :babelast:`decorator`
    """
    type = "Decorator"
    expression = attr(converter=Call.converter)


@cls
class ClassProperty(AST, DecorateMixin):
    """
    :babelast:`classproperty`
    """
    type = "ClassProperty"
    key = attr(converter=convert_as(Id), default=None)
    value = attr(converter=JS, default=None)
    computed = attr(default=False)
    static = attr(default=False)
    decorators = list_attr(converter=convert_as(Decorator))

    def as_method(self, **kwargs):
        """
        Convert class property to class method.
        """
        return ClassMethod(key=self.key, decorators=self.decorators, **kwargs)


@cls
class ClassMethod(AST, DecorateMixin):
    """
    :babelast:`classmethod`
    """
    type = "ClassMethod"
    key = attr(converter=Identifier.auto, default=None)
    params = list_attr(converter=Identifier.auto)
    body = attr(default=None, converter=Block.converter)
    decorators = list_attr(converter=Decorator.converter)
    kind = attr(default=None)
    static = attr(default=False)
    generator = attr(default=False)
    isAsync = attr(default=False)
    id = attr(default=None)
    directives = list_attr(default=[])

    def as_property(self, **kwargs):
        """
        Convert class method to value-less class property.
        """
        return ClassProperty(key=self.key, decorators=self.decorators, **kwargs)

    @classmethod
    def getter(_cls, key, body, **kwargs): #pylint: disable=bad-classmethod-argument
        """
        Getter factory
        """
        kwargs['kind'] = 'get'
        return _cls(key, [], body, **kwargs)

    @classmethod
    def setter(_cls, key, body, **kwargs): #pylint: disable=bad-classmethod-argument
        """
        Setter factory
        """
        return _cls(key, [Id('val')], body, **kwargs)



@cls
class ClassBody(AST):
    """
    :babelast:`classbody`
    """
    type = "ClassBody"
    body = list_attr()


@cls
class Return(AST):
    """
    :babelast:`returnstatement`
    """
    type = "ReturnStatement"
    argument = attr(default=None)


@cls
class This(AST):
    """
    :babelast:`thisexpression`
    """
    type = "ThisExpression"

    def member(self, arg):
        """
        Get a member identifier
        """
        return Member(object=self, prop=Id.auto(arg))

This = This() #pylint: disable=invalid-name


@cls
class Class(AST):
    """
    :babelast:`classdeclaration`
    """
    type = "ClassDeclaration"
    id = attr(converter=Identifier.auto, default=None)
    superClass = attr(converter=Identifier.auto, default=None)
    body = attr(converter=convert_as(ClassBody),
                default=Factory(ClassBody))

    def add_property(self, key, val, decorators=None):
        """
        Append a property to class object
        """
        #pylint: disable=no-member
        decorators = decorators or []
        self.body.body.append(ClassProperty(key, val, decorators=decorators))


@cls
class ExportDefault(AST):
    """
    :babelast:`exportdefaultdeclaration`
    """
    type = "ExportDefaultDeclaration"
    declaration = attr(converter=convert_as(Identifier))



"""
`jast` utility methods
"""
import json
import copy

from types import GeneratorType
from functools import partial

from collections import OrderedDict
from cattr import Converter

from inflection import camelize as camelize, underscore
from attr import s as cls, ib as attr, Factory


class classproperty(object): #pylint: disable=invalid-name,too-few-public-methods
    """
    SO development ftw, https://stackoverflow.com/a/3203659
    """
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def is_list(val):
    """
    Iterable type inspection
    """
    list_types = [list, set, tuple, GeneratorType]
    return any([isinstance(val, x) for x in list_types])


def list_or_list(value):
    """
    Enforce list result
    """
    if is_list(value):
        return value
    return [value]


def is_string(val):
    """
    String type inspection
    """
    return isinstance(val, basestring)


class AliasFactory(Factory): #pylint: disable=too-few-public-methods
    """
    Generate a takes_self=True factory which returns a
    local attribute as factory predicate
    """
    def __init__(self, key, **kwargs):
        _factory = lambda x: getattr(x, key)
        takes_self = kwargs.pop('takes_self', True)
        super(AliasFactory, self).__init__(_factory, takes_self)


class ASTList(list):
    """
    :class:`.AST` mixin that produces list ast results
    by iterating local object.
    """
    def ast(self):
        """
        Return js AST representation
        """
        return [x.ast() for x in self]


class ConvList(ASTList):
    """
    :class:`.ASTList` object which additionally aplies converter to
    list items.
    """
    def __init__(self, items=None, convert=None):
        identity = lambda x: x
        self.convert = convert or identity
        items = items or []
        if not is_list(items):
            items = [items]

        items = list(items)
        items = map(self.convert, items)
        super(ConvList, self).__init__(items)

    def extend(self, arr):
        """
        Convert and delegate to list object
        """
        return list.append(self, map(self.convert, arr))

    def append(self, value):
        """
        Convert and delegate to list object
        """
        return list.append(self, self.convert(value))

    def insert(self, index, value):
        """
        Convert and delegate to list object
        """
        return list.insert(self, index, self.convert(value))

    def __setitem__(self, index, value):
        """
        Convert and delegate to list object
        """
        return list.__setitem__(self, index, self.convert(value))

    @classmethod
    def converter(_cls, _type): #pylint: disable=bad-classmethod-argument
        """
        Expose as attribute converter
        """
        return partial(ConvList, convert=_type)


def convert_as(_type, expand_args=False):
    """
    Generate a converter the provided AST class.
    """
    def _converter(value):
        """
        Introspects attribute value type to prevent double conversion.
        """
        if isinstance(value, _type):
            return value
        if expand_args:
            return _type(*value)
        return _type(value)
    return _converter


def list_attr(*args, **kwargs):
    """
    Declares :class:`.AST` object attribute as :class:`.ConvList`
    """
    convert = kwargs.get('converter', lambda x: x)
    factory = partial(ConvList, *args, convert=convert)
    return attr(default=Factory(factory), converter=factory)


def js_var(name, camelize_first=False):
    """
    Common :class:`.Identifier` converter.
    """
    if isinstance(name, basestring):
        from jast.babel import Identifier #pylint: disable=cyclic-import
        if '.' in name:
            return Identifier.auto(name)
        elif '-' in name:
            return camelize(underscore(name),
                            uppercase_first_letter=camelize_first)
        elif camelize_first:
            return camelize(name)
    return name


class _ASTConverter(Converter):
    """
    :class:`cattr.Converter` which applies a bit of ast magic.
    """
    def unstructure_attrs_asdict(self, obj):
        attrs = super(_ASTConverter, self).unstructure_attrs_asdict(obj)

        if hasattr(obj, 'ast_map'):
            mapping = obj.ast_map
            for key, val in mapping.items():
                attrs[val] = attrs[key]
                del attrs[key]
        if hasattr(obj, 'type'):
            attrs['type'] = obj.type
        if hasattr(obj, 'extra_ast_attrs'):
            extra = obj.extra_ast_attrs
            dispatch = self._unstructure_func.dispatch
            for name in extra:
                var = getattr(obj, name)
                attrs[name] = dispatch(var)(var)
        ordering = {'type': -100}
        attrs_order_map = getattr(obj, 'ast_keys_order',
                                  [x.name for x in obj.__attrs_attrs__])
        ordering.update(dict([(x, attrs_order_map.index(x)) for x in attrs_order_map]))
        ordered_items = sorted(attrs.items(), key=lambda x: ordering.get(x[0], 1))
        return OrderedDict(ordered_items)


@cls
class AST(object):
    """
    Base of all AST structure classes.
    """

    @classproperty
    def converter(self):
        """
        Expose a converter for the underlying AST structure.
        """
        return convert_as(self)

    def copy(self):
        """
        Make a clone of current AST instance.
        """
        return copy.deepcopy(self)

    @staticmethod
    def collect_definitions(ast):
        """
        Walk through AST structure and collect identifier definition entries. Expose
        them as AST :class:`.Import` statements.
        """
        imports = []
        ast = copy.deepcopy(ast)
        if not ast:
            return ast, imports

        if isinstance(ast, list):
            for i, value in enumerate(ast):
                val, _imp = AST.collect_definitions(value)
                ast[i] = val
                imports += _imp

        if isinstance(ast, dict):
            for key, value in ast.items():
                val, _imp = AST.collect_definitions(value)
                ast[key] = val
                imports += _imp

            if 'type' in ast and ast.get('type') in ['Identifier', 'MemberExpression']:
                if ast.get('definition'):
                    imports.append(ast.get('definition'))
                if 'definition' in ast:
                    del ast['definition']
        return ast, imports

    def ast(self):
        """
        Expose current AST instance to AST representation (dict or list).
        """
        converter = _ASTConverter()
        data = converter.unstructure(self)
        return data

    def ast_defs(self):
        """
        Helper methods to parse current AST instance as ast/imports tuple.
        """
        return AST.collect_definitions(self.ast())

    def json(self):
        """
        JSON representation of underlying AST instance
        """
        ast, _ = AST.collect_definitions(self.ast())
        return json.dumps(ast, indent=4)

    def to_dict(self):
        return json.loads(self.json())

    def json_defs(self):
        """
        JSON representation of :class:`.Import` statements related to current AST instance.
        """
        ast, imports = AST.collect_definitions(self.ast())
        return json.dumps(imports, indent=4), json.dumps(ast, indent=4)

    @classmethod
    def partial(_cls, *args, **kwargs): #pylint: disable=bad-classmethod-argument
        """
        Helper util to expose current AST structure as a partial constructor.
        """
        return partial(cls, *args, **kwargs)


@cls #pylint: disable=too-few-public-methods
class DecorateMixin(object):
    """
    AST mixin for structrures which can decorate other AST instances.
    """

    def decorate(self, item):
        """
        Expose current AST instance as a decorator
        """
        from jast.babel import Decorator #pylint: disable=cyclic-import

        decorator = self
        if not isinstance(decorator, Decorator):
            decorator = Decorator(self)
        assert hasattr(item, 'decorators')
        item.decorators.append(decorator)
        return self


def parse_module(mod):
    """
    Collect AST objects from a python module
    """
    keys = dir(mod)
    assert 'AST' in keys
    asts = getattr(mod, 'AST')
    if type(asts) != dict:
        asts = {
            mod.__name__: asts
        }
    return asts

def batch(*files):
    """
    Expose a set of :class:`jast.custom.File` objects into a grouped json ast structure.
    """
    data = {}
    for astfile in files:
        data[astfile.name] = astfile.ast()
    return json.dumps(data, indent=4)



"""
jast module
"""
from .babel import \
    Program, \
    Identifier, \
    JS, \
    String, \
    Numeric, \
    Boolean, \
    Import, \
    ImportSpecifier, \
    ImportDefaultSpecifier, \
    Object, \
    ObjectProperty, \
    Null, \
    Array, \
    Member, \
    Var, \
    VariableDeclarator, \
    Expression, \
    Block, \
    Function, \
    NamedFunction, \
    Call, \
    Decorator, \
    CallDecorator, \
    Class, \
    ClassProperty, \
    ClassMethod, \
    ClassBody, \
    Return, \
    This, \
    ExportDefault

from .custom import Code, File
from .util import batch

__all__ = [
    'Program', 'Identifier', 'JS', 'String', 'Numeric', 'Boolean',
    'Import', 'ImportSpecifier', 'ImportDefaultSpecifier', 'Object',
    'ObjectProperty', 'Null', 'Array', 'Member', 'Var', 'VariableDeclarator',
    'Expression', 'Block', 'Function', 'NamedFunction', 'Call', 'Decorator',
    'CallDecorator', 'Class', 'ClassProperty', 'ClassMethod', 'ClassBody',
    'Return', 'This', 'ExportDefault', 'Code', 'File',
    'batch'
]



"""
`jast` custom AST structures
"""
from jast.util import list_attr, cls, convert_as, attr, json
from jast.babel import AST, ExportDefault, Program


@cls
class Code(AST):
    """
    Custom AST type to represent a string of raw javascript code.
    """
    type = "Code"
    code = attr()


@cls
class File(AST):
    """
    Custom jast AST helper structure.
    """
    name = attr()
    body = list_attr()
    export = attr(default=None, converter=convert_as(ExportDefault))
    imports = list_attr()
    header = list_attr()

    def ast(self):
        #pylint: disable=not-an-iterable,no-member
        prog = Program()
        for imp in self.imports:
            prog.body.append(imp)
        for header in self.header:
            prog.body.append(header)
        for stm in self.body:
            prog.body.append(stm)

        export_converter = convert_as(ExportDefault)
        self.export = export_converter(self.export)
        if self.export and self.export.declaration.name:
            prog.body.append(self.export)

        ast, defs = prog.ast_defs()
        imports = [x.get('type') == 'ImportDeclaration' for x in defs]
        other = [x.get('type') != 'ImportDeclaration' for x in  defs]

        for decl in other:
            ast['body'].insert(0, decl)

        imported = []
        imports.reverse()
        for decl in imports:
            key = json.dumps(decl)
            if key in imported:
                continue
            imported.append(key)
            ast['body'].insert(0, decl)

        # collect identifiers imports
        return {
            "type": "File",
            "program": ast
        }
